#!/usr/bin/env python3

import json
import logging
import os
import pathlib
import time

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

log = logging.getLogger("NP.test")
log.setLevel("DEBUG")
log.parent.setLevel("WARNING")

try:
    from plotly_resampler import register_plotly_resampler, unregister_plotly_resampler

    plotly_resampler_installed = True
except ImportError:
    plotly_resampler_installed = False
    log.error("Importing plotly failed. Interactive plots will not work.")

from neuralprophet import NeuralProphet, set_random_seed

DIR = pathlib.Path(__file__).parent.parent.absolute()
DATA_DIR = os.path.join(DIR, "tests", "test-data")
PEYTON_FILE = os.path.join(DATA_DIR, "wp_log_peyton_manning.csv")
AIR_FILE = os.path.join(DATA_DIR, "air_passengers.csv")
YOS_FILE = os.path.join(DATA_DIR, "yosemite_temps.csv")
ENERGY_PRICE_DAILY_FILE = os.path.join(DATA_DIR, "tutorial04_kaggle_energy_daily_temperature.csv")

# Important to set seed for reproducibility
set_random_seed(42)


def get_system_speed():
    repeats = 5
    benchmarks = np.array([])
    for a in range(0, repeats):
        start = time.time()
        for i in range(0, 1000):
            for x in range(1, 1000):
                3.141592 * 2**x
            for x in range(1, 1000):
                float(x) / 3.141592
            for x in range(1, 1000):
                float(3.141592) / x

        end = time.time()
        duration = end - start
        duration = round(duration, 3)
        benchmarks = np.append(benchmarks, duration)

    log.info(f"System speed: {round(np.mean(benchmarks), 5)}s")
    log.info(f"Standart deviation: {round(np.std(benchmarks), 5)}s")
    return benchmarks.mean(), benchmarks.std()


def create_metrics_plot(metrics):
    # Deactivate the resampler since it is not compatible with kaleido (image export)
    if plotly_resampler_installed:
        unregister_plotly_resampler()

    # Plotly params
    prediction_color = "#2d92ff"
    actual_color = "black"
    line_width = 2
    xaxis_args = {"showline": True, "mirror": True, "linewidth": 1.5, "showgrid": False}
    yaxis_args = {
        "showline": True,
        "mirror": True,
        "linewidth": 1.5,
        "showgrid": False,
        "rangemode": "tozero",
        "type": "log",
    }
    layout_args = {
        "autosize": True,
        "template": "plotly_white",
        "margin": go.layout.Margin(l=0, r=10, b=0, t=30, pad=0),
        "font": dict(size=10),
        "title": dict(font=dict(size=10)),
        "width": 1000,
        "height": 200,
    }

    metric_cols = [col for col in metrics.columns if not ("_val" in col or col == "RegLoss" or col == "epoch")]
    fig = make_subplots(rows=1, cols=len(metric_cols), subplot_titles=metric_cols)
    for i, metric in enumerate(metric_cols):
        fig.add_trace(
            go.Scatter(
                y=metrics[metric],
                name=metric,
                mode="lines",
                line=dict(color=prediction_color, width=line_width),
                legendgroup=metric,
            ),
            row=1,
            col=i + 1,
        )
        if f"{metric}_val" in metrics.columns:
            fig.add_trace(
                go.Scatter(
                    y=metrics[f"{metric}_val"],
                    name=f"{metric}_val",
                    mode="lines",
                    line=dict(color=actual_color, width=line_width),
                    legendgroup=metric,
                ),
                row=1,
                col=i + 1,
            )
        if metric == "Loss":
            fig.add_trace(
                go.Scatter(
                    y=metrics["RegLoss"],
                    name="RegLoss",
                    mode="lines",
                    line=dict(color=actual_color, width=line_width),
                    legendgroup=metric,
                ),
                row=1,
                col=i + 1,
            )
    fig.update_xaxes(xaxis_args)
    fig.update_yaxes(yaxis_args)
    fig.update_layout(layout_args)
    return fig


def test_PeytonManning():
    df = pd.read_csv(PEYTON_FILE)
    m = NeuralProphet()
    df_train, df_test = m.split_df(df=df, freq="D", valid_p=0.1)

    system_speed, std = get_system_speed()
    start = time.time()
    metrics = m.fit(df_train, validation_df=df_test, freq="D")  # , early_stopping=True)
    end = time.time()

    accuracy_metrics = metrics.to_dict("records")[-1]
    accuracy_metrics["time"] = round(end - start, 2)
    accuracy_metrics["system_performance"] = round(system_speed, 5)
    accuracy_metrics["system_std"] = round(std, 5)
    with open(os.path.join(DIR, "tests", "metrics", "PeytonManning.json"), "w") as outfile:
        json.dump(accuracy_metrics, outfile)

    create_metrics_plot(metrics).write_image(os.path.join(DIR, "tests", "metrics", "PeytonManning.svg"))


def test_YosemiteTemps():
    df = pd.read_csv(YOS_FILE)
    m = NeuralProphet(
        n_lags=36,
        n_forecasts=12,
        changepoints_range=0.9,
        n_changepoints=30,
        weekly_seasonality=False,
    )
    df_train, df_test = m.split_df(df=df, freq="5min", valid_p=0.1)

    system_speed, std = get_system_speed()
    start = time.time()
    metrics = m.fit(df_train, validation_df=df_test, freq="5min")  # , early_stopping=True)
    end = time.time()

    accuracy_metrics = metrics.to_dict("records")[-1]
    accuracy_metrics["time"] = round(end - start, 2)
    accuracy_metrics["system_performance"] = round(system_speed, 5)
    accuracy_metrics["system_std"] = round(std, 5)
    with open(os.path.join(DIR, "tests", "metrics", "YosemiteTemps.json"), "w") as outfile:
        json.dump(accuracy_metrics, outfile)

    create_metrics_plot(metrics).write_image(os.path.join(DIR, "tests", "metrics", "YosemiteTemps.svg"))


def test_AirPassengers():
    df = pd.read_csv(AIR_FILE)
    m = NeuralProphet(seasonality_mode="multiplicative")
    df_train, df_test = m.split_df(df=df, freq="MS", valid_p=0.1)

    system_speed, std = get_system_speed()
    start = time.time()
    metrics = m.fit(df_train, validation_df=df_test, freq="MS")  # , early_stopping=True)
    end = time.time()

    accuracy_metrics = metrics.to_dict("records")[-1]
    accuracy_metrics["time"] = round(end - start, 2)
    accuracy_metrics["system_performance"] = round(system_speed, 5)
    accuracy_metrics["system_std"] = round(std, 5)
    with open(os.path.join(DIR, "tests", "metrics", "AirPassengers.json"), "w") as outfile:
        json.dump(accuracy_metrics, outfile)

    create_metrics_plot(metrics).write_image(os.path.join(DIR, "tests", "metrics", "AirPassengers.svg"))


def test_EnergyPriceDaily():
    df = pd.read_csv(ENERGY_PRICE_DAILY_FILE)
    df["temp"] = df["temperature"]

    m = NeuralProphet(
        n_forecasts=7,
        n_changepoints=0,
        yearly_seasonality=True,
        weekly_seasonality=True,
        daily_seasonality=False,
        n_lags=14,
    )
    m.add_lagged_regressor("temp", n_lags=3)
    m.add_future_regressor("temperature")

    df_train, df_test = m.split_df(df=df, freq="D", valid_p=0.1)

    system_speed, std = get_system_speed()
    start = time.time()
    metrics = m.fit(df_train, validation_df=df_test, freq="D")  # , early_stopping=True)
    end = time.time()

    accuracy_metrics = metrics.to_dict("records")[-1]
    accuracy_metrics["time"] = round(end - start, 2)
    accuracy_metrics["system_performance"] = round(system_speed, 5)
    accuracy_metrics["system_std"] = round(std, 5)
    with open(os.path.join(DIR, "tests", "metrics", "EnergyPriceDaily.json"), "w") as outfile:
        json.dump(accuracy_metrics, outfile)

    create_metrics_plot(metrics).write_image(os.path.join(DIR, "tests", "metrics", "EnergyPriceDaily.svg"))



from abc import abstractmethod

import torch
import torch.nn as nn

from neuralprophet import utils
from neuralprophet.components.seasonality import Seasonality
from neuralprophet.utils_torch import init_parameter


class FourierSeasonality(Seasonality):
    def __init__(
        self,
        config,
        id_list,
        quantiles,
        num_seasonalities_modelled,
        num_seasonalities_modelled_dict,
        n_forecasts,
        device,
    ):
        super().__init__(
            config=config,
            n_forecasts=n_forecasts,
            num_seasonalities_modelled=num_seasonalities_modelled,
            num_seasonalities_modelled_dict=num_seasonalities_modelled_dict,
            quantiles=quantiles,
            id_list=id_list,
            device=device,
        )
        self.season_dims = utils.config_seasonality_to_model_dims(self.config_seasonality)
        if self.season_dims is not None:
            # Seasonality parameters for global or local modelling
            self.season_params = nn.ParameterDict(
                {
                    # dimensions -
                    # [no. of quantiles, num_seasonalities_modelled, no. of fourier terms for each seasonality]
                    name: init_parameter(
                        dims=[len(self.quantiles)] + [self.num_seasonalities_modelled_dict[name]] + [dim]
                    )
                    for name, dim in self.season_dims.items()
                }
            )

    @abstractmethod
    def compute_fourier(self, features, name, meta=None):
        """Compute single seasonality component.

        Parameters
        ----------
            features : torch.Tensor, float
                Features related to seasonality component, dims: (batch, n_forecasts, n_features)
            name : str
                Name of seasonality. for attribution to corresponding model weights.
            meta: dict
                Metadata about the all the samples of the model input batch. Contains the following:
                    * ``df_name`` (list, str), time series ID corresponding to each sample of the input batch.

        Returns
        -------
            torch.Tensor
                Forecast component of dims (batch, n_forecasts)
        """
        pass

    def forward(self, s, meta):
        """Compute all seasonality components.

        Parameters
        ----------
            s : torch.Tensor, float
                dict of named seasonalities (keys) with their features (values)
                dims of each dict value (batch, n_forecasts, n_features)
            meta: dict
                Metadata about the all the samples of the model input batch. Contains the following:
                    * ``df_name`` (list, str), time series ID corresponding to each sample of the input batch.

        Returns
        -------
            torch.Tensor
                Forecast component of dims (batch, n_forecasts)
        """
        device = s[list(s.keys())[0]].device
        x = torch.zeros(
            size=(s[list(s.keys())[0]].shape[0], s[list(s.keys())[0]].shape[1], len(self.quantiles)),
            device=device,
        )
        for name, features in s.items():
            x = x + self.compute_fourier(features, name, meta)
        return x


class GlobalFourierSeasonality(FourierSeasonality):
    def __init__(
        self,
        config,
        id_list,
        quantiles,
        num_seasonalities_modelled,
        num_seasonalities_modelled_dict,
        n_forecasts,
        device,
    ):
        super().__init__(
            config=config,
            n_forecasts=n_forecasts,
            num_seasonalities_modelled=num_seasonalities_modelled,
            num_seasonalities_modelled_dict=num_seasonalities_modelled_dict,
            quantiles=quantiles,
            id_list=id_list,
            device=device,
        )

    def compute_fourier(self, features, name, meta=None):
        """Compute single seasonality component.

        Parameters
        ----------
            features : torch.Tensor, float
                Features related to seasonality component, dims: (batch, n_forecasts, n_features)
            name : str
                Name of seasonality. for attribution to corresponding model weights.
            meta: dict
                Metadata about the all the samples of the model input batch. Contains the following:
                    * ``df_name`` (list, str), time series ID corresponding to each sample of the input batch.

        Returns
        -------
            torch.Tensor
                Forecast component of dims (batch, n_forecasts)
        """
        # dimensions -  batch_size, n_forecasts, quantiles
        seasonality = torch.sum(
            features.unsqueeze(dim=2) * self.season_params[name].permute(1, 0, 2).unsqueeze(dim=0), dim=-1
        )
        return seasonality


class LocalFourierSeasonality(FourierSeasonality):
    def __init__(
        self,
        config,
        id_list,
        quantiles,
        num_seasonalities_modelled,
        num_seasonalities_modelled_dict,
        n_forecasts,
        device,
    ):
        super().__init__(
            config=config,
            n_forecasts=n_forecasts,
            num_seasonalities_modelled=num_seasonalities_modelled,
            num_seasonalities_modelled_dict=num_seasonalities_modelled_dict,
            quantiles=quantiles,
            id_list=id_list,
            device=device,
        )

    def compute_fourier(self, features, name, meta=None):
        """Compute single seasonality component.

        Parameters
        ----------
            features : torch.Tensor, float
                Features related to seasonality component, dims: (batch, n_forecasts, n_features)
            name : str
                Name of seasonality. for attribution to corresponding model weights.
            meta: dict
                Metadata about the all the samples of the model input batch. Contains the following:
                    * ``df_name`` (list, str), time series ID corresponding to each sample of the input batch.

        Returns
        -------
            torch.Tensor
                Forecast component of dims (batch, n_forecasts)
        """
        # From the dataloader meta data, we get the one-hot encoding of the df_name.
        meta_name_tensor_one_hot = nn.functional.one_hot(meta, num_classes=len(self.id_list))
        # dimensions - quantiles, batch, parameters_fourier
        season_params_sample = torch.sum(
            meta_name_tensor_one_hot.unsqueeze(dim=0).unsqueeze(dim=-1) * self.season_params[name].unsqueeze(dim=1),
            dim=2,
        )
        # dimensions -  batch_size, n_forecasts, quantiles
        seasonality = torch.sum(features.unsqueeze(2) * season_params_sample.permute(1, 0, 2).unsqueeze(1), dim=-1)
        return seasonality


class GlocalFourierSeasonality(FourierSeasonality):
    def __init__(
        self,
        config,
        id_list,
        quantiles,
        num_seasonalities_modelled,
        num_seasonalities_modelled_dict,
        n_forecasts,
        device,
    ):
        super().__init__(
            config=config,
            n_forecasts=n_forecasts,
            num_seasonalities_modelled=num_seasonalities_modelled,
            num_seasonalities_modelled_dict=num_seasonalities_modelled_dict,
            quantiles=quantiles,
            id_list=id_list,
            device=device,
        )

    def compute_fourier(self, features, name, meta=None):
        """Compute single seasonality component.

        Parameters
        ----------
            features : torch.Tensor, float
                Features related to seasonality component, dims: (batch, n_forecasts, n_features)
            name : str
                Name of seasonality. for attribution to corresponding model weights.
            meta: dict
                Metadata about the all the samples of the model input batch. Contains the following:
                    * ``df_name`` (list, str), time series ID corresponding to each sample of the input batch.

        Returns
        -------
            torch.Tensor
                Forecast component of dims (batch, n_forecasts)
        """
        # From the dataloader meta data, we get the one-hot encoding of the df_name.
        if self.config_seasonality.periods[name].global_local == "local":
            meta_name_tensor_one_hot = nn.functional.one_hot(meta, num_classes=len(self.id_list))
            # dimensions - quantiles, batch, parameters_fourier
            season_params_sample = torch.sum(
                meta_name_tensor_one_hot.unsqueeze(dim=0).unsqueeze(dim=-1) * self.season_params[name].unsqueeze(dim=1),
                dim=2,
            )
            # dimensions -  batch_size, n_forecasts, quantiles
            seasonality = torch.sum(features.unsqueeze(2) * season_params_sample.permute(1, 0, 2).unsqueeze(1), dim=-1)
        elif self.config_seasonality.periods[name].global_local == "global":
            # dimensions -  batch_size, n_forecasts, quantiles
            seasonality = torch.sum(
                features.unsqueeze(dim=2) * self.season_params[name].permute(1, 0, 2).unsqueeze(dim=0), dim=-1
            )
        return seasonality



import logging

import pandas as pd

from neuralprophet import df_utils
from neuralprophet.configure import Normalization

log = logging.getLogger("NP.data.transforming")


def _normalize(df: pd.DataFrame, config_normalization: Normalization) -> pd.DataFrame:
    """Apply data scales.

    Applies data scaling factors to df using data_params.

    Parameters
    ----------
        df : pd.DataFrame
            dataframe containing column ``ds``, ``y``, and optionally``ID`` with all data
        config_normalization: Normalization
            Normalization configuration

    Returns
    -------
        df: pd.DataFrame, normalized
    """
    df, _, _, _ = df_utils.prep_or_copy_df(df)
    df_norm = pd.DataFrame()
    for df_name, df_i in df.groupby("ID"):
        data_params = config_normalization.get_data_params(df_name)
        df_i.drop("ID", axis=1, inplace=True)
        df_aux = df_utils.normalize(df_i, data_params).copy(deep=True)
        df_aux["ID"] = df_name
        df_norm = pd.concat((df_norm, df_aux), ignore_index=True)
    return df_norm



# https://pytorch-lightning.readthedocs.io/en/stable/extensions/logging.html#make-a-custom-logger
import collections
from typing import Any, Mapping, Optional

import pytorch_lightning as pl
from pytorch_lightning.callbacks import TQDMProgressBar
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.utilities.rank_zero import rank_zero_only


class MetricsLogger(TensorBoardLogger):
    def __init__(
        self,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.history = collections.defaultdict(list)
        self.checkpoint_path = None

    def after_save_checkpoint(self, checkpoint_callback) -> None:
        """Called after model checkpoint callback saves a new checkpoint.

        Args:
            checkpoint_callback: the model checkpoint callback instance
        """
        self.checkpoint_path = checkpoint_callback.best_model_path

    @rank_zero_only
    def log_metrics(self, metrics: Mapping[str, float], step: Optional[int] = None) -> None:
        super(MetricsLogger, self).log_metrics(metrics, step)
        # metrics is a dictionary of metric names and values
        for metric_name, metric_value in metrics.items():
            if metric_name == "hp_metric":
                pass
            elif metric_name != "epoch":
                self.history[metric_name].append(metric_value)
            else:  # case epoch. We want to avoid adding multiple times the same. It happens for multiple losses.
                if (
                    not len(self.history["epoch"]) or not self.history["epoch"][-1] == metric_value  # len == 0:
                ):  # the last values of epochs is not the one we are currently trying to add.
                    self.history["epoch"].append(metric_value)
                else:
                    pass
        return


class ProgressBar(TQDMProgressBar):
    """
    Custom progress bar for PyTorch Lightning for only update every epoch, not every batch.
    """

    def __init__(self, *args, **kwargs):
        self.epochs = kwargs.pop("epochs")
        super().__init__(*args, **kwargs)

    def on_train_epoch_start(self, trainer: "pl.Trainer", *_) -> None:
        self.main_progress_bar.reset(self.epochs)
        self.main_progress_bar.set_description(f"Epoch {trainer.current_epoch + 1}")
        self._update_n(self.main_progress_bar, trainer.current_epoch + 1)

    def on_train_batch_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", *_) -> None:
        pass

    def _update_n(self, bar, value: int) -> None:
        if not bar.disable:
            bar.n = value
            bar.refresh()



#!/usr/bin/env python3

import logging
import os
import pathlib

import pandas as pd
import pytest

from neuralprophet import NeuralProphet, uncertainty_evaluate

log = logging.getLogger("NP.test")
log.setLevel("DEBUG")
log.parent.setLevel("WARNING")

DIR = pathlib.Path(__file__).parent.parent.absolute()
DATA_DIR = os.path.join(DIR, "tests", "test-data")
PEYTON_FILE = os.path.join(DATA_DIR, "wp_log_peyton_manning.csv")
AIR_FILE = os.path.join(DATA_DIR, "air_passengers.csv")
YOS_FILE = os.path.join(DATA_DIR, "yosemite_temps.csv")
NROWS = 256
EPOCHS = 1
BATCH_SIZE = 128
LR = 1.0


def test_uncertainty_estimation_peyton_manning():
    log.info("testing: Uncertainty Estimation Peyton Manning")
    df = pd.read_csv(PEYTON_FILE)
    playoffs = pd.DataFrame(
        {
            "event": "playoff",
            "ds": pd.to_datetime(
                [
                    "2008-01-13",
                    "2009-01-03",
                    "2010-01-16",
                    "2010-01-24",
                    "2010-02-07",
                    "2011-01-08",
                    "2013-01-12",
                    "2014-01-12",
                    "2014-01-19",
                    "2014-02-02",
                    "2015-01-11",
                    "2016-01-17",
                    "2016-01-24",
                    "2016-02-07",
                ]
            ),
        }
    )
    superbowls = pd.DataFrame(
        {
            "event": "superbowl",
            "ds": pd.to_datetime(["2010-02-07", "2014-02-02", "2016-02-07"]),
        }
    )
    events_df = pd.concat((playoffs, superbowls))

    m = NeuralProphet(
        n_forecasts=1,
        loss_func="SmoothL1Loss",
        quantiles=[0.01, 0.99],
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
    )

    # add lagged regressors
    if m.n_lags > 0:
        df["A"] = df["y"].rolling(7, min_periods=1).mean()
        df["B"] = df["y"].rolling(30, min_periods=1).mean()
        m = m.add_lagged_regressor(name="A")
        m = m.add_lagged_regressor(name="B", only_last_value=True)

    # add events
    m = m.add_events(["superbowl", "playoff"], lower_window=-1, upper_window=1, regularization=0.1)

    m = m.add_country_holidays("US", mode="additive", regularization=0.1)

    df["C"] = df["y"].rolling(7, min_periods=1).mean()
    df["D"] = df["y"].rolling(30, min_periods=1).mean()

    m = m.add_future_regressor(name="C", regularization=0.1)
    m = m.add_future_regressor(name="D", regularization=0.1)

    history_df = m.create_df_with_events(df, events_df)

    m.fit(history_df, freq="D")

    periods = 90
    regressors_future_df = pd.DataFrame(data={"C": df["C"][:periods], "D": df["D"][:periods]})
    future_df = m.make_future_dataframe(
        df=history_df,
        regressors_df=regressors_future_df,
        events_df=events_df,
        periods=periods,
        n_historic_predictions=360,
    )
    m.predict(df=future_df)


def test_uncertainty_estimation_yosemite_temps():
    log.info("testing: Uncertainty Estimation Yosemite Temps")
    df = pd.read_csv(YOS_FILE)
    m = NeuralProphet(
        n_lags=12,
        n_forecasts=6,
        quantiles=[0.01, 0.99],
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
    )

    m.fit(df, freq="5min")
    future = m.make_future_dataframe(df, periods=6, n_historic_predictions=3 * 24 * 12)
    m.predict(future)
    m.highlight_nth_step_ahead_of_each_forecast(m.n_forecasts)


def test_uncertainty_estimation_air_travel():
    log.info("testing: Uncertainty Estimation Air Travel")
    df = pd.read_csv(AIR_FILE)
    m = NeuralProphet(
        seasonality_mode="multiplicative",
        loss_func="MSE",
        quantiles=[0.01, 0.99],
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
    )
    m.fit(df, freq="MS")
    future = m.make_future_dataframe(df, periods=50, n_historic_predictions=len(df))
    m.predict(future)


def test_uncertainty_estimation_multiple_quantiles():
    log.info("testing: Uncertainty Estimation Air Travel")
    df = pd.read_csv(AIR_FILE)
    multi_quantiles = [
        [0.5],  # forecast shows only yhat1, no duplicate yhat1 50.0%
        [0.8],  # forecast yhat1 and yhat1 80.0%
        [0.3, 0.6, 0.9],
        [0.05, 0.25, 0.75, 0.95],
    ]
    for quantiles in multi_quantiles:
        m = NeuralProphet(
            seasonality_mode="multiplicative",
            loss_func="MSE",
            quantiles=quantiles,
            epochs=EPOCHS,
            batch_size=BATCH_SIZE,
            learning_rate=LR,
        )
        m.fit(df, freq="MS")
        future = m.make_future_dataframe(df, periods=50, n_historic_predictions=len(df))
        m.predict(future)


def test_split_conformal_prediction():
    log.info("testing: Naive Split Conformal Prediction Air Travel")
    df = pd.read_csv(AIR_FILE)
    m = NeuralProphet(
        seasonality_mode="multiplicative",
        loss_func="MSE",
        quantiles=[0.05, 0.95],
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
    )

    train_df, test_df = m.split_df(df, freq="MS", valid_p=0.2)
    train_df, cal_df = m.split_df(train_df, freq="MS", valid_p=0.15)
    m.fit(train_df, freq="MS")

    alpha = 0.1
    decompose = False
    for method in ["naive", "cqr"]:  # Naive and CQR SCP methods
        for show_all_PI in [True, False]:
            future = m.make_future_dataframe(
                test_df,
                periods=50,
                n_historic_predictions=len(test_df),
            )
            forecast = m.conformal_predict(
                future,
                calibration_df=cal_df,
                alpha=alpha,
                method=method,
                decompose=decompose,
                show_all_PI=show_all_PI,
            )
            uncertainty_evaluate(forecast)


def test_asymmetrical_quantiles():
    log.info(
        "testing: Naive Split Conformal Prediction and Conformalized Quantile Regression " "with asymmetrical quantiles"
    )
    df = pd.read_csv(AIR_FILE)
    m = NeuralProphet(
        seasonality_mode="multiplicative",
        loss_func="MSE",
        quantiles=[0.05, 0.95],
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
    )

    train_df, test_df = m.split_df(df, freq="MS", valid_p=0.2)
    train_df, cal_df = m.split_df(train_df, freq="MS", valid_p=0.15)
    m.fit(train_df, freq="MS")

    alpha = (0.03, 0.07)
    decompose = False
    future = m.make_future_dataframe(
        test_df,
        periods=50,
        n_historic_predictions=len(test_df),
    )

    # should raise value error if method is naive and alpha is not a float
    method = "naive"
    with pytest.raises(ValueError):
        forecast = m.conformal_predict(
            future,
            calibration_df=cal_df,
            alpha=alpha,
            method=method,
            decompose=decompose,
        )

    # should not raise value error if method is cqr and alpha is not a float
    method = "cqr"
    forecast = m.conformal_predict(
        future,
        calibration_df=cal_df,
        alpha=alpha,
        method=method,
        decompose=decompose,
    )
    uncertainty_evaluate(forecast)



#!/usr/bin/env python3

import logging
import os
import pathlib

import pandas as pd

from neuralprophet import TorchProphet as Prophet

log = logging.getLogger("NP.test")
log.setLevel("DEBUG")
log.parent.setLevel("WARNING")

DIR = pathlib.Path(__file__).parent.parent.absolute()
DATA_DIR = os.path.join(DIR, "tests", "test-data")
PEYTON_FILE = os.path.join(DATA_DIR, "wp_log_peyton_manning.csv")
AIR_FILE = os.path.join(DATA_DIR, "air_passengers.csv")
NROWS = 256
EPOCHS = 1
BATCH_SIZE = 128
LR = 1.0

PLOT = False


def test_wrapper_base():
    log.info("testing: Wrapper base")
    df = pd.read_csv(PEYTON_FILE, nrows=NROWS)
    m = Prophet(epochs=EPOCHS, batch_size=BATCH_SIZE, learning_rate=LR)
    m.fit(df)

    future = m.make_future_dataframe(periods=50)
    m.predict(future)


def test_wrapper_components():
    log.info("testing: Wrapper components")
    df = pd.read_csv(AIR_FILE, nrows=NROWS)
    df["regressor"] = df["y"].rolling(7, min_periods=1).mean()
    regressors_df_future = pd.DataFrame(data={"regressor": df["regressor"][-50:]})

    m = Prophet(
        seasonality_mode="multiplicative",
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
    )

    m.add_seasonality("quarterly", period=91.25, fourier_order=8, mode="additive")
    m.add_regressor("regressor", mode="additive")
    m.fit(df)

    future = m.make_future_dataframe(periods=50, regressors_df=regressors_df_future)
    m.predict(future)


def test_wrapper_warnings():
    log.info("testing: Wrapper base")
    pd.read_csv(PEYTON_FILE, nrows=NROWS)
    # Args should be handled by raising warnings
    Prophet(
        seasonality_prior_scale=0.0,
        mcmc_samples=0,
        stan_backend="CMDSTANPY",
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
    )


def test_wrapper_plots():
    log.info("testing: Wrapper plots")
    df = pd.read_csv(PEYTON_FILE, nrows=NROWS)
    m = Prophet(epochs=EPOCHS, batch_size=BATCH_SIZE, learning_rate=LR)
    m.fit(df)

    future = m.make_future_dataframe(periods=50)
    forecast = m.predict(future)

    fig1 = m.plot(forecast)
    fig2 = m.plot(forecast, plotting_backend="plotly")
    fig3 = m.plot_components(forecast)
    fig4 = m.plot_components(forecast, plotting_backend="plotly")

    if PLOT:
        fig1.show()
        fig2.show()
        fig3.show()
        fig4.show()



import pytest

from neuralprophet.configure import Train


def generate_config_train_params(overrides={}):
    config_train_params = {
        "quantiles": None,
        "learning_rate": None,
        "epochs": None,
        "batch_size": None,
        "loss_func": "SmoothL1Loss",
        "optimizer": "AdamW",
    }
    for key, value in overrides.items():
        config_train_params[key] = value
    return config_train_params


def test_config_training_quantiles():
    checks = [
        ({}, [0.5]),
        ({"quantiles": None}, [0.5]),
        ({"quantiles": []}, [0.5]),
        ({"quantiles": [0.2]}, [0.5, 0.2]),
        ({"quantiles": [0.2, 0.8]}, [0.5, 0.2, 0.8]),
        ({"quantiles": [0.5, 0.8]}, [0.5, 0.8]),
    ]

    for overrides, expected in checks:
        config_train_params = generate_config_train_params(overrides)
        config = Train(**config_train_params)
        assert config.quantiles == expected


def test_config_training_quantiles_error_invalid_type():
    config_train_params = generate_config_train_params()
    config_train_params["quantiles"] = "hello world"
    with pytest.raises(AssertionError) as err:
        Train(**config_train_params)
    assert str(err.value) == "Quantiles must be in a list format, not None or scalar."


def test_config_training_quantiles_error_invalid_scale():
    config_train_params = generate_config_train_params()
    config_train_params["quantiles"] = [-1]
    with pytest.raises(Exception) as err:
        Train(**config_train_params)
    assert str(err.value) == "The quantiles specified need to be floats in-between (0, 1)."



import pytest

from neuralprophet.__main__ import parse_args
from neuralprophet._version import __version__


def test_main_file(capsys):
    with pytest.raises(SystemExit) as exit_info:
        parse_args(["--version"])

    out, _ = capsys.readouterr()
    assert exit_info.value.code == 0
    assert __version__ in out



import logging

# matplotlib is known to spam logs, thus we set the logging level to warning
logging.getLogger("matplotlib").setLevel(logging.WARNING)
logging.getLogger("pytorch_lightning").setLevel(logging.INFO)
logging.getLogger("torch").setLevel(logging.WARNING)
logging.getLogger("fsspec").setLevel(logging.INFO)



import logging
import math
from collections import OrderedDict
from functools import reduce
from typing import Dict, List, Optional, Union

import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn as nn
import torchmetrics

from neuralprophet import configure, np_types
from neuralprophet.components.router import get_future_regressors, get_seasonality, get_trend
from neuralprophet.utils import (
    check_for_regularization,
    config_events_to_model_dims,
    reg_func_events,
    reg_func_regressors,
    reg_func_season,
    reg_func_seasonality_glocal,
    reg_func_trend,
    reg_func_trend_glocal,
)
from neuralprophet.utils_torch import init_parameter, interprete_model

log = logging.getLogger("NP.time_net")


class TimeNet(pl.LightningModule):
    """Linear time regression fun and some not so linear fun.
    A modular model that models classic time-series components
        * trend
        * seasonality
        * auto-regression (as AR-Net)
        * covariates (as AR-Net)
        * apriori regressors
        * events and holidays
    by using Neural Network components.
    The Auto-regression and covariate components can be configured as a deeper network (AR-Net).
    """

    def __init__(
        self,
        config_seasonality: configure.ConfigSeasonality,
        config_train: Optional[configure.Train] = None,
        config_trend: Optional[configure.Trend] = None,
        config_ar: Optional[configure.AR] = None,
        config_normalization: Optional[configure.Normalization] = None,
        config_lagged_regressors: Optional[configure.ConfigLaggedRegressors] = None,
        config_regressors: Optional[configure.ConfigFutureRegressors] = None,
        config_events: Optional[configure.ConfigEvents] = None,
        config_holidays: Optional[configure.ConfigCountryHolidays] = None,
        n_forecasts: int = 1,
        n_lags: int = 0,
        max_lags: int = 0,
        ar_layers: Optional[List[int]] = [],
        lagged_reg_layers: Optional[List[int]] = [],
        compute_components_flag: bool = False,
        metrics: Optional[np_types.CollectMetricsMode] = {},
        id_list: List[str] = ["__df__"],
        num_trends_modelled: int = 1,
        num_seasonalities_modelled: int = 1,
        num_seasonalities_modelled_dict: dict = None,
        meta_used_in_model: bool = False,
    ):
        """
        Parameters
        ----------
            quantiles : list
                the set of quantiles estimated
            config_train : configure.Train
            config_trend : configure.Trend
            config_seasonality : configure.ConfigSeasonality
            config_ar : configure.AR
            config_lagged_regressors : configure.ConfigLaggedRegressors
                Configurations for lagged regressors
            config_regressors : configure.ConfigFutureRegressors
                Configs of regressors with mode and index.
            config_events : configure.ConfigEvents
            config_holidays : OrderedDict
            config_normalization: OrderedDict
            n_forecasts : int
                number of steps to forecast. Aka number of model outputs
            n_lags : int
                number of previous steps of time series used as input (aka AR-order)
                Note
                ----
                The default value is ``0``, which initializes no auto-regression.

            max_lags : int
                Number of max. previous steps of time series used as input (aka AR-order).

            ar_layers : list
                List of hidden layers (for AR-Net).

                Note
                ----
                The default value is ``[]``, which initializes no hidden layers.

            lagged_reg_layers : list
                List of hidden layers (for covariate-Net).

                Note
                ----
                The default value is ``[]``, which initializes no hidden layers.


            compute_components_flag : bool
                Flag whether to compute the components of the model or not.
            metrics : dict
                Dictionary of torchmetrics to be used during training and for evaluation.
            id_list : list
                List of different time series IDs, used for global-local modelling (if enabled)
                Note
                ----
                This parameter is set to  ``['__df__']`` if only one time series is input.
            num_trends_modelled : int
                Number of different trends modelled.
                Note
                ----
                If only 1 time series is modelled, it will be always 1.
                Note
                ----
                For multiple time series. If trend is modelled globally the value is set
                to 1, otherwise it is set to the number of time series modelled.
            num_seasonalities_modelled : int
                Number of different seasonalities modelled.
                Note
                ----
                If only 1 time series is modelled, it will be always 1.
                Note
                ----
                For multiple time series. If seasonality is modelled globally the value is set
                to 1, otherwise it is set to the number of time series modelled.
            meta_used_in_model : boolean
                Whether we need to know the time series ID when we interact with the Model.
                Note
                ----
                Will be set to ``True`` if more than one component is modelled locally.
        """
        super().__init__()

        # Store hyerparameters in model checkpoint
        # TODO: causes a RuntimeError under certain conditions, investigate and handle better
        try:
            self.save_hyperparameters()
        except RuntimeError:
            pass

        # General
        self.n_forecasts = n_forecasts

        # Lightning Config
        self.config_train = config_train
        self.config_normalization = config_normalization
        self.compute_components_flag = compute_components_flag

        # Optimizer and LR Scheduler
        self._optimizer = self.config_train.optimizer
        self._scheduler = self.config_train.scheduler
        self.automatic_optimization = False

        # Hyperparameters (can be tuned using trainer.tune())
        self.learning_rate = self.config_train.learning_rate if self.config_train.learning_rate is not None else 1e-3
        self.batch_size = self.config_train.batch_size

        # Metrics Config
        self.metrics_enabled = bool(metrics)  # yields True if metrics is not an empty dictionary
        if self.metrics_enabled:
            metrics = {metric: torchmetrics.__dict__[metrics[metric][0]](**metrics[metric][1]) for metric in metrics}
            self.log_args = {
                "on_step": False,
                "on_epoch": True,
                "prog_bar": True,
                "batch_size": self.config_train.batch_size,
            }
            self.metrics_train = torchmetrics.MetricCollection(metrics=metrics)
            self.metrics_val = torchmetrics.MetricCollection(metrics=metrics, postfix="_val")

        # For Multiple Time Series Analysis
        self.id_list = id_list
        self.id_dict = dict((key, i) for i, key in enumerate(id_list))
        self.num_trends_modelled = num_trends_modelled
        self.num_seasonalities_modelled = num_seasonalities_modelled
        self.num_seasonalities_modelled_dict = num_seasonalities_modelled_dict
        self.meta_used_in_model = meta_used_in_model

        # Regularization
        self.reg_enabled = check_for_regularization(
            [
                config_seasonality,
                config_regressors,
                config_lagged_regressors,
                config_ar,
                config_events,
                config_trend,
                config_holidays,
            ]
        )

        # Quantiles
        self.quantiles = self.config_train.quantiles

        # Trend
        self.config_trend = config_trend
        self.trend = get_trend(
            config=config_trend,
            id_list=id_list,
            quantiles=self.quantiles,
            num_trends_modelled=num_trends_modelled,
            n_forecasts=n_forecasts,
            device=self.device,
        )

        # Seasonalities
        self.config_seasonality = config_seasonality
        # Error handling
        if self.config_seasonality is not None:
            if self.config_seasonality.mode == "multiplicative" and self.config_trend is None:
                raise ValueError("Multiplicative seasonality requires trend.")
            if self.config_seasonality.mode not in ["additive", "multiplicative"]:
                raise ValueError(f"Seasonality Mode {self.config_seasonality.mode} not implemented.")
            # Initialize seasonality
            self.seasonality = get_seasonality(
                config=config_seasonality,
                id_list=id_list,
                quantiles=self.quantiles,
                num_seasonalities_modelled=num_seasonalities_modelled,
                num_seasonalities_modelled_dict=num_seasonalities_modelled_dict,
                n_forecasts=n_forecasts,
                device=self.device,
            )

        # Events
        self.config_events = config_events
        self.config_holidays = config_holidays
        self.events_dims = config_events_to_model_dims(self.config_events, self.config_holidays)
        if self.events_dims is not None:
            n_additive_event_params = 0
            n_multiplicative_event_params = 0
            for event, configs in self.events_dims.items():
                if configs["mode"] not in ["additive", "multiplicative"]:
                    log.error("Event Mode {} not implemented. Defaulting to 'additive'.".format(configs["mode"]))
                    self.events_dims[event]["mode"] = "additive"
                if configs["mode"] == "additive":
                    n_additive_event_params += len(configs["event_indices"])
                elif configs["mode"] == "multiplicative":
                    if self.config_trend is None:
                        log.error("Multiplicative events require trend.")
                        raise ValueError
                    n_multiplicative_event_params += len(configs["event_indices"])
            self.event_params = nn.ParameterDict(
                {
                    # dimensions - [no. of quantiles, no. of additive events]
                    "additive": init_parameter(dims=[len(self.quantiles), n_additive_event_params]),
                    # dimensions - [no. of quantiles, no. of multiplicative events]
                    "multiplicative": init_parameter(dims=[len(self.quantiles), n_multiplicative_event_params]),
                }
            )
        else:
            self.config_events = None
            self.config_holidays = None

        # Autoregression
        self.config_ar = config_ar
        self.n_lags = n_lags
        self.ar_layers = ar_layers
        self.max_lags = max_lags
        if self.n_lags > 0:
            self.ar_net = nn.ModuleList()
            d_inputs = self.n_lags
            for d_hidden_i in self.ar_layers:
                self.ar_net.append(nn.Linear(d_inputs, d_hidden_i, bias=True))
                d_inputs = d_hidden_i
            # final layer has input size d_inputs and output size equal to no. of forecasts * no. of quantiles
            self.ar_net.append(nn.Linear(d_inputs, self.n_forecasts * len(self.quantiles), bias=False))
            for lay in self.ar_net:
                nn.init.kaiming_normal_(lay.weight, mode="fan_in")

        # Lagged regressors
        self.lagged_reg_layers = lagged_reg_layers
        self.config_lagged_regressors = config_lagged_regressors
        if self.config_lagged_regressors is not None:
            self.covar_net = nn.ModuleList()
            d_inputs = sum([covar.n_lags for _, covar in self.config_lagged_regressors.items()])
            for d_hidden_i in self.lagged_reg_layers:
                self.covar_net.append(nn.Linear(d_inputs, d_hidden_i, bias=True))
                d_inputs = d_hidden_i
            self.covar_net.append(nn.Linear(d_inputs, self.n_forecasts * len(self.quantiles), bias=False))
            for lay in self.covar_net:
                nn.init.kaiming_normal_(lay.weight, mode="fan_in")

        # Regressors
        self.config_regressors = config_regressors
        if self.config_regressors.regressors is not None:
            # Initialize future_regressors
            self.future_regressors = get_future_regressors(
                config=config_regressors,
                id_list=id_list,
                quantiles=self.quantiles,
                n_forecasts=n_forecasts,
                device=self.device,
                config_trend_none_bool=self.config_trend is None,
            )
        else:
            self.config_regressors.regressors = None

    @property
    def ar_weights(self) -> torch.Tensor:
        """sets property auto-regression weights for regularization. Update if AR is modelled differently"""
        # TODO: this is wrong for deep networks, use utils_torch.interprete_model
        return self.ar_net[0].weight

    def get_covar_weights(self, covar_input=None) -> torch.Tensor:
        """
        Get attributions of covariates network w.r.t. the model input.
        """
        if self.config_lagged_regressors is not None:
            # Accumulate the lags of the covariates
            covar_splits = np.add.accumulate(
                [covar.n_lags for _, covar in self.config_lagged_regressors.items()][:-1]
            ).tolist()
            # If actual covariates are provided, use them to compute the attributions
            if covar_input is not None:
                covar_input = torch.cat([covar for _, covar in covar_input.items()], axis=1)
            # Calculate the attributions w.r.t. the inputs
            if self.lagged_reg_layers == []:
                attributions = self.covar_net[0].weight
            else:
                attributions = interprete_model(self, "covar_net", "forward_covar_net", covar_input)
            # Split the attributions into the different covariates
            attributions_split = torch.tensor_split(
                attributions,
                covar_splits,
                axis=1,
            )
            # Combine attributions and covariate name
            covar_attributions = dict(zip(self.config_lagged_regressors.keys(), attributions_split))
        else:
            covar_attributions = None
        return covar_attributions

    def set_covar_weights(self, covar_weights: torch.Tensor):
        """
        Function to set the covariate weights for later interpretation in compute_components.
        This function is needed since the gradient information is not available during the predict_step
        method and attributions cannot be calculated in compute_components.

        :param covar_weights: _description_
        :type covar_weights: torch.Tensor
        """
        self.covar_weights = covar_weights

    def get_event_weights(self, name: str) -> Dict[str, torch.Tensor]:
        """
        Retrieve the weights of event features given the name
        Parameters
        ----------
            name : str
                Event name
        Returns
        -------
            OrderedDict
                Dict of the weights of all offsets corresponding to a particular event
        """

        event_dims = self.events_dims[name]
        mode = event_dims["mode"]

        if mode == "multiplicative":
            event_params = self.event_params["multiplicative"]
        else:
            assert mode == "additive"
            event_params = self.event_params["additive"]

        event_param_dict = OrderedDict({})
        for event_delim, indices in zip(event_dims["event_delim"], event_dims["event_indices"]):
            event_param_dict[event_delim] = event_params[:, indices : (indices + 1)]
        return event_param_dict

    def _compute_quantile_forecasts_from_diffs(self, diffs: torch.Tensor, predict_mode: bool = False) -> torch.Tensor:
        """
        Computes the actual quantile forecasts from quantile differences estimated from the model
        Args:
            diffs : torch.Tensor
                tensor of dims (batch, n_forecasts, no_quantiles) which
                contains the median quantile forecasts as well as the diffs of other quantiles
                from the median quantile
            predict_mode : bool
                boolean variable indicating whether the model is in prediction mode
        Returns:
            dim (batch, n_forecasts, no_quantiles)
                final forecasts
        """
        if len(self.quantiles) > 1:
            # generate the actual quantile forecasts from predicted differences
            if any(quantile > 0.5 for quantile in self.quantiles):
                quantiles_divider_index = next(i for i, quantile in enumerate(self.quantiles) if quantile > 0.5)
            else:
                quantiles_divider_index = len(self.quantiles)

            n_upper_quantiles = diffs.shape[-1] - quantiles_divider_index
            n_lower_quantiles = quantiles_divider_index - 1

            out = torch.zeros_like(diffs)
            out[:, :, 0] = diffs[:, :, 0]  # set the median where 0 is the median quantile index

            if n_upper_quantiles > 0:  # check if upper quantiles exist
                upper_quantile_diffs = diffs[:, :, quantiles_divider_index:]
                if predict_mode:  # check for quantile crossing and correct them in predict mode
                    upper_quantile_diffs[:, :, 0] = torch.max(
                        torch.tensor(0, device=self.device), upper_quantile_diffs[:, :, 0]
                    )
                    for i in range(n_upper_quantiles - 1):
                        next_diff = upper_quantile_diffs[:, :, i + 1]
                        diff = upper_quantile_diffs[:, :, i]
                        upper_quantile_diffs[:, :, i + 1] = torch.max(next_diff, diff)
                out[:, :, quantiles_divider_index:] = (
                    upper_quantile_diffs + diffs[:, :, 0].unsqueeze(dim=2).repeat(1, 1, n_upper_quantiles).detach()
                )  # set the upper quantiles

            if n_lower_quantiles > 0:  # check if lower quantiles exist
                lower_quantile_diffs = diffs[:, :, 1:quantiles_divider_index]
                if predict_mode:  # check for quantile crossing and correct them in predict mode
                    lower_quantile_diffs[:, :, -1] = torch.max(
                        torch.tensor(0, device=self.device), lower_quantile_diffs[:, :, -1]
                    )
                    for i in range(n_lower_quantiles - 1, 0, -1):
                        next_diff = lower_quantile_diffs[:, :, i - 1]
                        diff = lower_quantile_diffs[:, :, i]
                        lower_quantile_diffs[:, :, i - 1] = torch.max(next_diff, diff)
                lower_quantile_diffs = -lower_quantile_diffs
                out[:, :, 1:quantiles_divider_index] = (
                    lower_quantile_diffs + diffs[:, :, 0].unsqueeze(dim=2).repeat(1, 1, n_lower_quantiles).detach()
                )  # set the lower quantiles
        else:
            out = diffs
        return out

    def scalar_features_effects(self, features: torch.Tensor, params: nn.Parameter, indices=None) -> torch.Tensor:
        """
        Computes events component of the model
        Parameters
        ----------
            features : torch.Tensor, float
                Features (either additive or multiplicative) related to event component dims (batch, n_forecasts,
                n_features)
            params : nn.Parameter
                Params (either additive or multiplicative) related to events
            indices : list of int
                Indices in the feature tensors related to a particular event
        Returns
        -------
            torch.Tensor
                Forecast component of dims (batch, n_forecasts)
        """
        if indices is not None:
            features = features[:, :, indices]
            params = params[:, indices]

        return torch.sum(features.unsqueeze(dim=2) * params.unsqueeze(dim=0).unsqueeze(dim=0), dim=-1)

    def auto_regression(self, lags: Union[torch.Tensor, float]) -> torch.Tensor:
        """Computes auto-regessive model component AR-Net.
        Parameters
        ----------
            lags  : torch.Tensor, float
                Previous times series values, dims: (batch, n_lags)
        Returns
        -------
            torch.Tensor
                Forecast component of dims: (batch, n_forecasts)
        """
        x = lags
        for i in range(len(self.ar_layers) + 1):
            if i > 0:
                x = nn.functional.relu(x)
            x = self.ar_net[i](x)

        # segment the last dimension to match the quantiles
        x = x.reshape(x.shape[0], self.n_forecasts, len(self.quantiles))
        return x

    def forward_covar_net(self, covariates):
        """Compute all covariate components.
        Parameters
        ----------
            covariates : dict(torch.Tensor, float)
                dict of named covariates (keys) with their features (values)
                dims of each dict value: (batch, n_lags)
        Returns
        -------
            torch.Tensor
                Forecast component of dims (batch, n_forecasts, quantiles)
        """
        # Concat covariates into one tensor)
        if isinstance(covariates, dict):
            x = torch.cat([covar for _, covar in covariates.items()], axis=1)
        else:
            x = covariates
        for i in range(len(self.lagged_reg_layers) + 1):
            if i > 0:
                x = nn.functional.relu(x)
            x = self.covar_net[i](x)

        # segment the last dimension to match the quantiles
        x = x.reshape(x.shape[0], self.n_forecasts, len(self.quantiles))
        return x

    def forward(self, inputs: Dict, meta: Dict = None, compute_components_flag: bool = False) -> torch.Tensor:
        """This method defines the model forward pass.
        Note
        ----
        Time input is required. Minimum model setup is a linear trend.
        Parameters
        ----------
            inputs : dict
                Model inputs, each of len(df) but with varying dimensions
                Note
                ----
                Contains the following data:
                Model Inputs
                    * ``time`` (torch.Tensor , loat), normalized time, dims: (batch, n_forecasts)
                    * ``lags`` (torch.Tensor, float), dims: (batch, n_lags)
                    * ``seasonalities`` (torch.Tensor, float), dict of named seasonalities (keys) with their features
                    (values), dims of each dict value (batch, n_forecasts, n_features)
                    * ``covariates`` (torch.Tensor, float), dict of named covariates (keys) with their features
                    (values), dims of each dict value: (batch, n_lags)
                    * ``events`` (torch.Tensor, float), all event features, dims (batch, n_forecasts, n_features)
                    * ``regressors``(torch.Tensor, float), all regressor features, dims (batch, n_forecasts, n_features)
                    * ``predict_mode`` (bool), optional and only passed during prediction
            meta : dict, default=None
                Metadata about the all the samples of the model input batch.
                Contains the following:
                Model Meta:
                    * ``df_name`` (list, str), time series ID corresponding to each sample of the input batch.
                Note
                ----
                The meta is sorted in the same way the inputs are sorted.
                Note
                ----
                The default None value allows the forward method to be used without providing the meta argument.
                This was designed to avoid issues with the library `lr_finder` https://github.com/davidtvs/pytorch-lr-finder
                while having  ``config_trend.trend_global_local="local"``.
                The turnaround consists on passing the same meta (dummy ID) to all the samples of the batch.
                Internally, this is equivalent to use ``config_trend.trend_global_local="global"`` to find the optimal
                learning rate.
            compute_components_flag : bool, default=False
                If True, components will be computed.

        Returns
        -------
            torch.Tensor
                Forecast of dims (batch, n_forecasts, no_quantiles)
        """
        # Turnaround to avoid issues when the meta argument is None and meta_used_in_model
        if meta is None and self.meta_used_in_model:
            name_id_dummy = self.id_list[0]
            meta = OrderedDict()
            meta["df_name"] = [name_id_dummy for _ in range(inputs["time"].shape[0])]
            meta = torch.tensor([self.id_dict[i] for i in meta["df_name"]], device=self.device)

        components = {}
        additive_components = torch.zeros(
            size=(inputs["time"].shape[0], self.n_forecasts, len(self.quantiles)),
            device=self.device,
        )
        additive_components_nonstationary = torch.zeros(
            size=(inputs["time"].shape[0], inputs["time"].shape[1], len(self.quantiles)), device=self.device
        )
        multiplicative_components_nonstationary = torch.zeros(
            size=(inputs["time"].shape[0], inputs["time"].shape[1], len(self.quantiles)), device=self.device
        )

        trend = self.trend(t=inputs["time"], meta=meta)
        components["trend"] = trend

        if "seasonalities" in inputs:
            s = self.seasonality(s=inputs["seasonalities"], meta=meta)
            if self.config_seasonality.mode == "additive":
                additive_components_nonstationary += s
            elif self.config_seasonality.mode == "multiplicative":
                multiplicative_components_nonstationary += s
            components["seasonalities"] = s

        if "events" in inputs:
            if "additive" in inputs["events"].keys():
                additive_events = self.scalar_features_effects(
                    inputs["events"]["additive"], self.event_params["additive"]
                )
                additive_components_nonstationary += additive_events
                components["additive_events"] = additive_events
            if "multiplicative" in inputs["events"].keys():
                multiplicative_events = self.scalar_features_effects(
                    inputs["events"]["multiplicative"], self.event_params["multiplicative"]
                )
                multiplicative_components_nonstationary += multiplicative_events
                components["multiplicative_events"] = multiplicative_events

        if "regressors" in inputs:
            if "additive" in inputs["regressors"].keys():
                additive_regressors = self.future_regressors(inputs["regressors"]["additive"], "additive")
                additive_components_nonstationary += additive_regressors
                components["additive_regressors"] = additive_regressors
            if "multiplicative" in inputs["regressors"].keys():
                multiplicative_regressors = self.future_regressors(
                    inputs["regressors"]["multiplicative"], "multiplicative"
                )
                multiplicative_components_nonstationary += multiplicative_regressors
                components["multiplicative_regressors"] = multiplicative_regressors

        # stationarized input
        if "lags" in inputs:
            # combinde all non-stationary components over AR input range
            nonstationary_components = (  # dimensions - [batch, n_lags, median quantile]
                trend[:, : self.n_lags, 0]
                + additive_components_nonstationary[:, : self.n_lags, 0]
                + trend[:, : self.n_lags, 0].detach() * multiplicative_components_nonstationary[:, : self.n_lags, 0]
            )
            stationarized_lags = inputs["lags"] - nonstationary_components
            lags = self.auto_regression(lags=stationarized_lags)
            additive_components += lags
            components["lags"] = lags

        if "covariates" in inputs:
            covariates = self.forward_covar_net(covariates=inputs["covariates"])
            additive_components += covariates
            components["covariates"] = covariates

        # combine all non-stationary components over forecast range
        predictions_nonstationary = (
            trend[:, self.n_lags : inputs["time"].shape[1], :]
            + additive_components_nonstationary[:, self.n_lags : inputs["time"].shape[1], :]
            + trend[:, self.n_lags : inputs["time"].shape[1], :].detach()
            * multiplicative_components_nonstationary[:, self.n_lags : inputs["time"].shape[1], :]
        )
        prediction = predictions_nonstationary + additive_components  # dimensions - [batch, n_forecasts, no_quantiles]

        # check for crossing quantiles and correct them here
        if "predict_mode" in inputs.keys() and inputs["predict_mode"]:
            predict_mode = True
        else:
            predict_mode = False
        prediction_with_quantiles = self._compute_quantile_forecasts_from_diffs(prediction, predict_mode)

        # component calculation
        if compute_components_flag:
            components = self.compute_components(inputs, components, meta)
        else:
            components = None

        return prediction_with_quantiles, components

    def compute_components(self, inputs: Dict, components_raw: Dict, meta: Dict) -> Dict:
        """This method returns the values of each model component.
        Note
        ----
        Time input is required. Minimum model setup is a linear trend.
        Parameters
        ----------
            inputs : dict
                Model inputs, each of len(df) but with varying dimensions
                Note
                ----
                Contains the following data:
                Model Inputs
                    * ``time`` (torch.Tensor , loat), normalized time, dims: (batch, n_forecasts)
                    * ``lags`` (torch.Tensor, float), dims: (batch, n_lags)
                    * ``seasonalities`` (torch.Tensor, float), dict of named seasonalities (keys) with their features
                    (values), dims of each dict value (batch, n_forecasts, n_features)
                    * ``covariates`` (torch.Tensor, float), dict of named covariates (keys) with their features
                    (values), dims of each dict value: (batch, n_lags)
                    * ``events`` (torch.Tensor, float), all event features, dims (batch, n_forecasts, n_features)
                    * ``regressors``(torch.Tensor, float), all regressor features, dims (batch, n_forecasts, n_features)
            components_raw : dict
                components to be computed
        -------
            dict
                Containing forecast coomponents with elements of dims (batch, n_forecasts)
        """
        components = {}

        components["trend"] = components_raw["trend"][:, self.n_lags : inputs["time"].shape[1], :]
        if self.config_trend is not None and "seasonalities" in inputs:
            for name, features in inputs["seasonalities"].items():
                components[f"season_{name}"] = self.seasonality.compute_fourier(
                    features=features[:, self.n_lags : inputs["time"].shape[1], :], name=name, meta=meta
                )
        if self.n_lags > 0 and "lags" in inputs:
            components["ar"] = components_raw["lags"]
        if self.config_lagged_regressors is not None and "covariates" in inputs:
            # Combined forward pass
            all_covariates = components_raw["covariates"]
            # Calculate the contribution of each covariate on each forecast
            covar_attributions = self.covar_weights
            # Sum the contributions of all covariates
            covar_attribution_sum_per_forecast = reduce(
                torch.add, [torch.sum(covar, axis=1) for _, covar in covar_attributions.items()]
            ).to(all_covariates.device)
            for name in inputs["covariates"].keys():
                # Distribute the contribution of the current covariate to the combined forward pass
                # 1. Calculate the relative share of each covariate on the total attributions
                # 2. Multiply the relative share with the combined forward pass
                components[f"lagged_regressor_{name}"] = torch.multiply(
                    all_covariates,
                    torch.divide(
                        torch.sum(covar_attributions[name], axis=1).to(all_covariates.device),
                        covar_attribution_sum_per_forecast,
                    ).reshape(self.n_forecasts, len(self.quantiles)),
                )
        if (self.config_events is not None or self.config_holidays is not None) and "events" in inputs:
            if "additive" in inputs["events"].keys():
                components["events_additive"] = components_raw["additive_events"][
                    :, self.n_lags : inputs["time"].shape[1], :
                ]
            if "multiplicative" in inputs["events"].keys():
                components["events_multiplicative"] = components_raw["multiplicative_events"][
                    :, self.n_lags : inputs["time"].shape[1], :
                ]
            for event, configs in self.events_dims.items():
                mode = configs["mode"]
                indices = configs["event_indices"]
                if mode == "additive":
                    features = inputs["events"]["additive"][:, self.n_lags : inputs["time"].shape[1], :]
                    params = self.event_params["additive"]
                else:
                    features = inputs["events"]["multiplicative"][:, self.n_lags : inputs["time"].shape[1], :]
                    params = self.event_params["multiplicative"]
                components[f"event_{event}"] = self.scalar_features_effects(
                    features=features, params=params, indices=indices
                )
        if self.config_regressors.regressors is not None and "regressors" in inputs:
            if "additive" in inputs["regressors"].keys():
                components["future_regressors_additive"] = components_raw["additive_regressors"][
                    :, self.n_lags : inputs["time"].shape[1], :
                ]

            if "multiplicative" in inputs["regressors"].keys():
                components["future_regressors_multiplicative"] = components_raw["multiplicative_regressors"][
                    :, self.n_lags : inputs["time"].shape[1], :
                ]

            for regressor, configs in self.future_regressors.regressors_dims.items():
                mode = configs["mode"]
                index = []
                index.append(configs["regressor_index"])
                features = inputs["regressors"][mode]
                components[f"future_regressor_{regressor}"] = self.future_regressors(
                    features[:, self.n_lags : inputs["time"].shape[1], :], mode, indeces=index
                )

        return components

    def set_compute_components(self, compute_components_flag):
        self.compute_components_flag = compute_components_flag

    def loss_func(self, inputs, predicted, targets):
        loss = None
        # Compute loss. no reduction.
        loss = self.config_train.loss_func(predicted, targets)
        # Weigh newer samples more.
        loss = loss * self._get_time_based_sample_weight(t=inputs["time"][:, self.n_lags :])
        loss = loss.sum(dim=2).mean()
        # Regularize.
        if self.reg_enabled:
            steps_per_epoch = math.ceil(self.trainer.estimated_stepping_batches / self.trainer.max_epochs)
            progress_in_epoch = 1 - ((steps_per_epoch * (self.current_epoch + 1) - self.global_step) / steps_per_epoch)
            loss, reg_loss = self._add_batch_regularizations(loss, self.current_epoch, progress_in_epoch)
        else:
            reg_loss = torch.tensor(0.0, device=self.device)
        return loss, reg_loss

    def training_step(self, batch, batch_idx):
        inputs, targets, meta = batch
        # Global-local
        if self.meta_used_in_model:
            meta_name_tensor = torch.tensor([self.id_dict[i] for i in meta["df_name"]], device=self.device)
        else:
            meta_name_tensor = None
        # Run forward calculation
        predicted, _ = self.forward(inputs, meta_name_tensor)
        # Store predictions in self for later network visualization
        self.train_epoch_prediction = predicted
        # Calculate loss
        loss, reg_loss = self.loss_func(inputs, predicted, targets)

        # Optimization
        optimizer = self.optimizers()
        optimizer.zero_grad()
        self.manual_backward(loss)
        optimizer.step()

        scheduler = self.lr_schedulers()
        scheduler.step()

        # Manually track the loss for the lr finder
        self.trainer.fit_loop.running_loss.append(loss)

        # Metrics
        if self.metrics_enabled:
            predicted_denorm = self.denormalize(predicted[:, :, 0])
            target_denorm = self.denormalize(targets.squeeze(dim=2))
            self.log_dict(self.metrics_train(predicted_denorm, target_denorm), **self.log_args)
            self.log("Loss", loss, **self.log_args)
            self.log("RegLoss", reg_loss, **self.log_args)
        return loss

    def validation_step(self, batch, batch_idx):
        inputs, targets, meta = batch
        # Global-local
        if self.meta_used_in_model:
            meta_name_tensor = torch.tensor([self.id_dict[i] for i in meta["df_name"]], device=self.device)
        else:
            meta_name_tensor = None
        # Run forward calculation
        predicted, _ = self.forward(inputs, meta_name_tensor)
        # Calculate loss
        loss, reg_loss = self.loss_func(inputs, predicted, targets)
        # Metrics
        if self.metrics_enabled:
            predicted_denorm = self.denormalize(predicted[:, :, 0])
            target_denorm = self.denormalize(targets.squeeze(dim=2))
            self.log_dict(self.metrics_val(predicted_denorm, target_denorm), **self.log_args)
            self.log("Loss_val", loss, **self.log_args)
            self.log("RegLoss_val", reg_loss, **self.log_args)

    def test_step(self, batch, batch_idx):
        inputs, targets, meta = batch
        # Global-local
        if self.meta_used_in_model:
            meta_name_tensor = torch.tensor([self.id_dict[i] for i in meta["df_name"]], device=self.device)
        else:
            meta_name_tensor = None
        # Run forward calculation
        predicted, _ = self.forward(inputs, meta_name_tensor)
        # Calculate loss
        loss, reg_loss = self.loss_func(inputs, predicted, targets)
        # Metrics
        if self.metrics_enabled:
            predicted_denorm = self.denormalize(predicted[:, :, 0])
            target_denorm = self.denormalize(targets.squeeze(dim=2))
            self.log_dict(self.metrics_val(predicted_denorm, target_denorm), **self.log_args)
            self.log("Loss_test", loss, **self.log_args)
            self.log("RegLoss_test", reg_loss, **self.log_args)

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        inputs, _, meta = batch
        # Global-local
        if self.meta_used_in_model:
            meta_name_tensor = torch.tensor([self.id_dict[i] for i in meta["df_name"]], device=self.device)
        else:
            meta_name_tensor = None
        # Add predict_mode flag to dataset
        inputs["predict_mode"] = True
        # Run forward calculation
        prediction, components = self.forward(inputs, meta_name_tensor, self.compute_components_flag)
        return prediction, components

    def configure_optimizers(self):
        # Optimizer
        optimizer = self._optimizer(self.parameters(), lr=self.learning_rate, **self.config_train.optimizer_args)

        # Scheduler
        lr_scheduler = self._scheduler(
            optimizer,
            max_lr=self.learning_rate,
            total_steps=self.trainer.estimated_stepping_batches,
            **self.config_train.scheduler_args,
        )

        return {"optimizer": optimizer, "lr_scheduler": lr_scheduler}

    def _get_time_based_sample_weight(self, t):
        weight = torch.ones_like(t)
        if self.config_train.newer_samples_weight > 1.0:
            end_w = self.config_train.newer_samples_weight
            start_t = self.config_train.newer_samples_start
            time = (t.detach() - start_t) / (1.0 - start_t)
            time = torch.maximum(torch.zeros_like(time), time)
            time = torch.minimum(torch.ones_like(time), time)  # time = 0 to 1
            time = np.pi * (time - 1.0)  # time =  -pi to 0
            time = 0.5 * torch.cos(time) + 0.5  # time =  0 to 1
            # scales end to be end weight times bigger than start weight
            # with end weight being 1.0
            weight = (1.0 + time * (end_w - 1.0)) / end_w
        return weight.unsqueeze(dim=2)  # add an extra dimension for the quantiles

    def _add_batch_regularizations(self, loss, epoch, progress):
        """Add regularization terms to loss, if applicable
        Parameters
        ----------
            loss : torch.Tensor, scalar
                current batch loss
            epoch : int
                current epoch number
            progress : float
                progress within the epoch, between 0 and 1
        Returns
        -------
            loss, reg_loss
        """
        delay_weight = self.config_train.get_reg_delay_weight(epoch, progress)

        reg_loss = torch.zeros(1, dtype=torch.float, requires_grad=False, device=self.device)
        if delay_weight > 0:
            # Add regularization of AR weights - sparsify
            if self.max_lags > 0 and self.config_ar.reg_lambda is not None:
                reg_ar = self.config_ar.regularize(self.ar_weights)
                reg_ar = torch.sum(reg_ar).squeeze() / self.n_forecasts
                reg_loss += self.config_ar.reg_lambda * reg_ar

            # Regularize trend to be smoother/sparse
            l_trend = self.config_trend.trend_reg
            if self.config_trend.n_changepoints > 0 and l_trend is not None and l_trend > 0:
                reg_trend = reg_func_trend(
                    weights=self.trend.get_trend_deltas,
                    threshold=self.config_train.trend_reg_threshold,
                )
                reg_loss += l_trend * reg_trend

            # Regularize seasonality: sparsify fourier term coefficients
            if self.config_seasonality:
                l_season = self.config_seasonality.reg_lambda
                if self.seasonality.season_dims is not None and l_season is not None and l_season > 0:
                    for name in self.seasonality.season_params.keys():
                        reg_season = reg_func_season(self.seasonality.season_params[name])
                        reg_loss += l_season * reg_season

            # Regularize events: sparsify events features coefficients
            if self.config_events is not None or self.config_holidays is not None:
                reg_events_loss = reg_func_events(self.config_events, self.config_holidays, self)
                reg_loss += reg_events_loss

            # Regularize regressors: sparsify regressor features coefficients
            if self.config_regressors.regressors is not None:
                reg_regressor_loss = reg_func_regressors(self.config_regressors.regressors, self)
                reg_loss += reg_regressor_loss

        trend_glocal_loss = torch.zeros(1, dtype=torch.float, requires_grad=False)
        # Glocal Trend
        if self.config_trend is not None:
            if self.config_trend.trend_global_local == "local" and self.config_trend.trend_local_reg != False:
                trend_glocal_loss = reg_func_trend_glocal(
                    self.trend.trend_k0, self.trend.trend_deltas, self.config_trend.trend_local_reg
                )
                reg_loss += trend_glocal_loss
        # Glocal Seasonality
        if self.config_seasonality is not None:
            if (
                self.config_seasonality.global_local in ["local", "glocal"]
                and self.config_seasonality.seasonality_local_reg != False
            ):
                seasonality_glocal_loss = reg_func_seasonality_glocal(
                    self.seasonality.season_params, self.config_seasonality.seasonality_local_reg
                )
                reg_loss += seasonality_glocal_loss
        loss = loss + reg_loss
        return loss, reg_loss

    def denormalize(self, ts):
        """
        Denormalize timeseries
        Parameters
        ----------
            target : torch.Tensor
                ts tensor
        Returns
        -------
            denormalized timeseries
        """
        if self.config_normalization.global_normalization:
            shift_y = (
                self.config_normalization.global_data_params["y"].shift
                if self.config_normalization.global_normalization and not self.config_normalization.normalize == "off"
                else 0
            )
            scale_y = (
                self.config_normalization.global_data_params["y"].scale
                if self.config_normalization.global_normalization and not self.config_normalization.normalize == "off"
                else 1
            )
            ts = scale_y * ts + shift_y
        return ts


class FlatNet(nn.Module):
    """
    Linear regression fun
    """

    def __init__(self, d_inputs, d_outputs):
        # Perform initialization of the pytorch superclass
        super(FlatNet, self).__init__()
        self.layers = nn.Sequential(
            nn.Linear(d_inputs, d_outputs),
        )
        nn.init.kaiming_normal_(self.layers[0].weight, mode="fan_in")

    def forward(self, x):
        return self.layers(x)

    @property
    def ar_weights(self):
        return self.model.layers[0].weight


class DeepNet(nn.Module):
    """
    A simple, general purpose, fully connected network
    """

    def __init__(self, d_inputs, d_outputs, lagged_reg_layers=[]):
        # Perform initialization of the pytorch superclass
        super(DeepNet, self).__init__()
        self.layers = nn.ModuleList()
        for d_hidden_i in lagged_reg_layers:
            self.layers.append(nn.Linear(d_inputs, d_hidden_i, bias=True))
            d_inputs = d_hidden_i
        self.layers.append(nn.Linear(d_inputs, d_outputs, bias=True))
        for lay in self.layers:
            nn.init.kaiming_normal_(lay.weight, mode="fan_in")

    def forward(self, x):
        """
        This method defines the network layering and activation functions
        """
        activation = nn.functional.relu
        for i in range(len(self.layers)):
            if i > 0:
                x = activation(x)
            x = self.layers[i](x)
        return x

    @property
    def ar_weights(self):
        return self.layers[0].weight



from neuralprophet.components import BaseComponent


class Seasonality(BaseComponent):
    def __init__(
        self,
        config,
        id_list,
        quantiles,
        num_seasonalities_modelled,
        num_seasonalities_modelled_dict,
        n_forecasts,
        device,
    ):
        super().__init__(n_forecasts=n_forecasts, quantiles=quantiles, id_list=id_list, device=device)
        self.config_seasonality = config
        self.num_seasonalities_modelled = num_seasonalities_modelled
        self.num_seasonalities_modelled_dict = num_seasonalities_modelled_dict



#!/usr/bin/env python3

import logging
import os
import pathlib

import pandas as pd

from neuralprophet import NeuralProphet

log = logging.getLogger("NP.test")
log.setLevel("DEBUG")
log.parent.setLevel("WARNING")

DIR = pathlib.Path(__file__).parent.parent.absolute()
DATA_DIR = os.path.join(DIR, "tests", "test-data")
PEYTON_FILE = os.path.join(DATA_DIR, "wp_log_peyton_manning.csv")
AIR_FILE = os.path.join(DATA_DIR, "air_passengers.csv")
YOS_FILE = os.path.join(DATA_DIR, "yosemite_temps.csv")
NROWS = 256
EPOCHS = 1
BATCH_SIZE = 128
LR = 1.0

PLOT = False


def test_trend_global_local_modeling():
    # TREND GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(
        n_forecasts=2, n_lags=10, epochs=EPOCHS, batch_size=BATCH_SIZE, learning_rate=LR, trend_global_local="local"
    )
    assert m.config_seasonality.global_local == "global"
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df)
    m.predict(future)
    m.test(test_df)
    m.predict_trend(test_df)
    m.predict_seasonal_components(test_df)
    m.plot_parameters()


def test_regularized_trend_global_local_modeling():
    # TREND GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(n_lags=10, epochs=EPOCHS, trend_global_local="local", trend_reg=1)
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df)
    m.predict(future)
    m.test(test_df)
    m.predict_trend(test_df)
    m.predict_seasonal_components(test_df)


def test_seasonality_global_local_modeling():
    # SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(
        n_forecasts=2, n_lags=10, epochs=EPOCHS, batch_size=BATCH_SIZE, learning_rate=LR, season_global_local="local"
    )
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df)
    m.predict(future)
    m.test(test_df)
    m.predict_trend(test_df)
    m.predict_seasonal_components(test_df)
    m.plot_parameters()


def test_changepoints0_global_local_modeling():
    # SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(
        n_forecasts=2,
        n_lags=10,
        n_changepoints=0,
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
        season_global_local="local",
    )
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df)
    m.predict(future)
    m.test(test_df)
    m.predict_trend(test_df)
    m.predict_seasonal_components(test_df)


def test_trend_discontinuous_global_local_modeling():
    # SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(
        n_forecasts=2,
        n_lags=10,
        growth="discontinuous",
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
        season_global_local="local",
    )
    assert m.config_trend.trend_global_local == "global"
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df)
    m.predict(future)
    m.test(test_df)
    m.predict_trend(test_df)
    m.predict_seasonal_components(test_df)


def test_attributes_global_local_modeling():
    # TREND GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(
        n_forecasts=2,
        n_lags=10,
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
        trend_global_local="local",
        season_global_local="local",
    )
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.1, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df)
    m.predict(future)
    assert "df1" in m.model.id_list
    assert m.model.num_trends_modelled == 3
    assert m.model.num_seasonalities_modelled == 3


def test_wrong_option_global_local_modeling():
    # SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(
        n_forecasts=2,
        n_lags=10,
        growth="discontinuous",
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
        season_global_local="glocsl",
        trend_global_local="glocsl",
    )
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df)
    forecast = m.predict(future)
    metrics = m.test(test_df)
    forecast_trend = m.predict_trend(test_df)
    forecast_seasonal_componets = m.predict_seasonal_components(test_df)


def test_different_seasonality_modeling():
    ### SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(
        n_forecasts=2,
        n_lags=10,
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
        season_global_local="local",
        yearly_seasonality_glocal_mode="global",
    )
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df)
    forecast = m.predict(future)
    metrics = m.test(test_df)
    forecast_trend = m.predict_trend(test_df)
    forecast_seasonal_componets = m.predict_seasonal_components(test_df)


def test_adding_new_global_seasonality():
    ### SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(
        n_forecasts=2,
        n_lags=10,
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
        season_global_local="local",
        yearly_seasonality_glocal_mode="global",
    )
    m.add_seasonality(period=30, fourier_order=8, name="monthly", global_local="global")
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df)
    forecast = m.predict(future)
    metrics = m.test(test_df)
    forecast_trend = m.predict_trend(test_df)
    forecast_seasonal_componets = m.predict_seasonal_components(test_df)


def test_adding_new_local_seasonality():
    ### SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(epochs=EPOCHS, batch_size=BATCH_SIZE, season_global_local="global", trend_global_local="local")
    m.add_seasonality(period=30, fourier_order=8, name="monthly", global_local="local")
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df, n_historic_predictions=True)
    forecast = m.predict(future)
    metrics = m.test(test_df)
    forecast_trend = m.predict_trend(test_df)
    forecast_seasonal_componets = m.predict_seasonal_components(test_df)


def test_trend_local_reg():
    ### SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    for coef_i in [-30, 0, False, True]:
        m = NeuralProphet(
            n_forecasts=1,
            epochs=EPOCHS,
            batch_size=BATCH_SIZE,
            learning_rate=LR,
            trend_global_local="local",
            trend_local_reg=coef_i,
        )

        m.add_seasonality(period=30, fourier_order=8, name="monthly", global_local="global")
        train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
        m.fit(train_df)
        future = m.make_future_dataframe(test_df, n_historic_predictions=True)
        forecast = m.predict(future)
        metrics = m.test(test_df)
        forecast_trend = m.predict_trend(test_df)
        forecast_seasonal_componets = m.predict_seasonal_components(test_df)


def test_glocal_seasonality_reg():
    ### SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    for coef_i in [-30, 0, False, True]:
        m = NeuralProphet(
            n_forecasts=1,
            epochs=EPOCHS,
            batch_size=BATCH_SIZE,
            learning_rate=LR,
            season_global_local="local",
            yearly_seasonality_glocal_mode="global",
            glocal_seasonality_reg=coef_i,
        )

        m.add_seasonality(period=30, fourier_order=8, name="monthly", global_local="global")
        train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
        m.fit(train_df)
        future = m.make_future_dataframe(test_df, n_historic_predictions=True)
        forecast = m.predict(future)
        metrics = m.test(test_df)


def test_trend_local_reg_if_global():
    ### SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    for coef_i in [-30, 0, False, True]:
        m = NeuralProphet(
            n_forecasts=1,
            epochs=EPOCHS,
            batch_size=BATCH_SIZE,
            learning_rate=LR,
            trend_global_local="global",
            trend_local_reg=3,
        )

        train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
        m.fit(train_df)
        future = m.make_future_dataframe(test_df, n_historic_predictions=True)
        forecast = m.predict(future)
        metrics = m.test(test_df)
        forecast_trend = m.predict_trend(test_df)
        forecast_seasonal_componets = m.predict_seasonal_components(test_df)


def test_different_seasonality_modeling():
    ### SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(
        n_forecasts=2,
        n_lags=10,
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
        season_global_local="local",
        yearly_seasonality_glocal_mode="global",
    )
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df)
    forecast = m.predict(future)
    metrics = m.test(test_df)
    forecast_trend = m.predict_trend(test_df)
    forecast_seasonal_componets = m.predict_seasonal_components(test_df)


def test_adding_new_global_seasonality():
    ### SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(
        n_forecasts=2,
        n_lags=10,
        epochs=EPOCHS,
        batch_size=BATCH_SIZE,
        learning_rate=LR,
        season_global_local="local",
        yearly_seasonality_glocal_mode="global",
    )
    m.add_seasonality(period=30, fourier_order=8, name="monthly", global_local="global")
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df)
    forecast = m.predict(future)
    metrics = m.test(test_df)
    forecast_trend = m.predict_trend(test_df)
    forecast_seasonal_componets = m.predict_seasonal_components(test_df)


def test_adding_new_local_seasonality():
    ### SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    m = NeuralProphet(epochs=EPOCHS, batch_size=BATCH_SIZE, season_global_local="global", trend_global_local="local")
    m.add_seasonality(period=30, fourier_order=8, name="monthly", global_local="local")
    train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
    m.fit(train_df)
    future = m.make_future_dataframe(test_df, n_historic_predictions=True)
    forecast = m.predict(future)
    metrics = m.test(test_df)
    forecast_trend = m.predict_trend(test_df)
    forecast_seasonal_componets = m.predict_seasonal_components(test_df)


def test_trend_local_reg():
    ### SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    for coef_i in [-30, 0, False, True]:
        m = NeuralProphet(
            n_forecasts=1,
            epochs=EPOCHS,
            batch_size=BATCH_SIZE,
            learning_rate=LR,
            trend_global_local="local",
            trend_local_reg=coef_i,
        )

        m.add_seasonality(period=30, fourier_order=8, name="monthly", global_local="global")
        train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
        m.fit(train_df)
        future = m.make_future_dataframe(test_df, n_historic_predictions=True)
        forecast = m.predict(future)
        metrics = m.test(test_df)
        forecast_trend = m.predict_trend(test_df)
        forecast_seasonal_componets = m.predict_seasonal_components(test_df)


def test_glocal_seasonality_reg():
    ### SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    for coef_i in [-30, 0, False, True]:
        m = NeuralProphet(
            n_forecasts=1,
            epochs=EPOCHS,
            batch_size=BATCH_SIZE,
            learning_rate=LR,
            season_global_local="local",
            yearly_seasonality_glocal_mode="global",
            seasonality_local_reg=coef_i,
        )

        m.add_seasonality(period=30, fourier_order=8, name="monthly", global_local="global")
        train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
        m.fit(train_df)
        future = m.make_future_dataframe(test_df, n_historic_predictions=True)
        forecast = m.predict(future)
        metrics = m.test(test_df)


def test_trend_local_reg_if_global():
    ### SEASONALITY GLOBAL LOCAL MODELLING - NO EXOGENOUS VARIABLES
    log.info("Global Modeling + Global Normalization")
    df = pd.read_csv(PEYTON_FILE, nrows=512)
    df1_0 = df.iloc[:128, :].copy(deep=True)
    df1_0["ID"] = "df1"
    df2_0 = df.iloc[128:256, :].copy(deep=True)
    df2_0["ID"] = "df2"
    df3_0 = df.iloc[256:384, :].copy(deep=True)
    df3_0["ID"] = "df3"
    for coef_i in [-30, 0, False, True]:
        m = NeuralProphet(
            n_forecasts=1,
            epochs=EPOCHS,
            batch_size=BATCH_SIZE,
            learning_rate=LR,
            trend_global_local="global",
            trend_local_reg=3,
        )

        train_df, test_df = m.split_df(pd.concat((df1_0, df2_0, df3_0)), valid_p=0.33, local_split=True)
        m.fit(train_df)
        future = m.make_future_dataframe(test_df, n_historic_predictions=True)
        forecast = m.predict(future)
        metrics = m.test(test_df)
        forecast_trend = m.predict_trend(test_df)
        forecast_seasonal_componets = m.predict_seasonal_components(test_df)



import numpy as np
import torch
import torch.nn as nn

from neuralprophet.components.trend import Trend
from neuralprophet.utils_torch import init_parameter


class PiecewiseLinearTrend(Trend):
    def __init__(self, config, id_list, quantiles, num_trends_modelled, n_forecasts, device):
        super().__init__(
            config=config,
            n_forecasts=n_forecasts,
            num_trends_modelled=num_trends_modelled,
            quantiles=quantiles,
            id_list=id_list,
            device=device,
        )

        self.segmentwise_trend = self.config_trend.trend_reg == 0

        # Trend_k0  parameter.
        # dimensions - [no. of quantiles,  num_trends_modelled, trend coeff shape]
        self.trend_k0 = init_parameter(dims=([len(self.quantiles)] + [self.num_trends_modelled] + [1]))

        # Configure the changepoints
        if self.config_trend.changepoints is None:
            # create equidistant changepoint times, including zero.
            linear_t = np.arange(self.config_trend.n_changepoints + 1).astype(float)
            linear_t = linear_t / (self.config_trend.n_changepoints + 1)
            self.config_trend.changepoints = self.config_trend.changepoints_range * linear_t
        else:
            self.config_trend.changepoints = np.insert(self.config_trend.changepoints, 0, 0.0)
        # Register in buffer so the tensor is moved to the correct device once initialized,
        # https://pytorch-lightning.readthedocs.io/en/stable/starter/converting.html#remove-any-cuda-or-to-device-calls
        self.register_buffer(
            "trend_changepoints_t",
            torch.tensor(self.config_trend.changepoints, requires_grad=False, dtype=torch.float),
        )

        # Trend Deltas parameters
        self.trend_deltas = init_parameter(
            dims=([len(self.quantiles)] + [self.num_trends_modelled] + [self.config_trend.n_changepoints + 1])
        )  # including first segment

        # When discontinuous, the start of the segment is not defined by the previous segments.
        # This brings a new set of parameters to optimize.
        if self.config_trend.growth == "discontinuous":
            self.trend_m = init_parameter(
                dims=([len(self.quantiles)] + [self.num_trends_modelled] + [self.config_trend.n_changepoints + 1])
            )  # including first segment

    def forward(self, t, meta):
        """Computes trend based on model configuration.

        Parameters
        ----------
            t : torch.Tensor float
                normalized time, dim: (batch, n_forecasts)
            meta: dict
                Metadata about the all the samples of the model input batch. Contains the following:
                    * ``df_name`` (list, str), time series ID corresponding to each sample of the input batch.
        Returns
        -------
            torch.Tensor
                Trend component, same dimensions as input t

        """
        # From the dataloader meta data, we get the one-hot encoding of the df_name.
        if self.config_trend.trend_global_local == "local":
            # dimensions - batch , num_time_series
            meta_name_tensor_one_hot = nn.functional.one_hot(meta, num_classes=len(self.id_list))
        else:
            meta_name_tensor_one_hot = None

        # Variables identifying, for t, the corresponding trend segment (for each sample of the batch).
        past_next_changepoint = t.unsqueeze(dim=2) >= self.trend_changepoints_t[1:].unsqueeze(dim=0)
        segment_id = past_next_changepoint.sum(dim=2)
        # = dimensions - batch_size, n_forecasts, segments (+ 1)
        current_segment = nn.functional.one_hot(segment_id, num_classes=self.config_trend.n_changepoints + 1)

        # Computing k_t.
        # dimensions - batch_size, n_forecasts, quantiles_size
        k_t = self.compute_k_t(current_segment, past_next_changepoint, meta_name_tensor_one_hot)

        # Computing m_t.
        # dimensions - batch_size, n_forecasts, quantiles_size
        m_t = self.compute_m_t(current_segment, past_next_changepoint, meta_name_tensor_one_hot)

        # Computing trend value at time(t) for each batch sample.
        # dimensions - batch_size, n_forecasts, quantiles_size
        trend = self.compute_trend(t, k_t, m_t, meta_name_tensor_one_hot)

        return self.bias.unsqueeze(dim=0).unsqueeze(dim=0) + trend

    @property
    def get_trend_deltas(self):
        """trend deltas for regularization.

        update if trend is modelled differently"""
        if self.config_trend is None or self.config_trend.n_changepoints < 1:
            trend_delta = None
        elif self.segmentwise_trend:
            trend_delta = self.trend_deltas[:, :, :] - torch.cat((self.trend_k0, self.trend_deltas[:, :, 0:-1]), dim=2)
        else:
            trend_delta = self.trend_deltas

        return trend_delta

    def add_regularization(self):
        pass

    def compute_k_t(self, current_segment, past_next_changepoint, meta_name_tensor_one_hot):
        """For segmentwise, k_t is the model parameter representing the trend slope(actually, trend slope-k_0) in the
        current_segment at time t (for each sample of the batch).

        For not segmentwise, k_t is the model parameter representing the difference between trend slope in the
        current_segment at time t and the trend slope in the previous segment (for each sample of the batch).

        Parameters
        ----------
            current_segment : torch.Tensor, int
                segment corresponding to time t (batch_size, n_forecasts, segments (+ 1))

            past_next_changepoint : torch.Tensor, bool
                whether the a changepoint >= time t (batch_size, n_forecasts, segments (+ 1))

            meta_name_tensor_one_hot : torch.Tensor, float
                Metadata about the all the samples of the model input batch.

                Contains the following:
                    * ``df_name`` (list, str), time series name ID corresponding to each sample of the input batch.
        Returns
        -------
            torch.Tensor
                k_t,  ( batch_size, n_forecasts, quantiles_size)
        """

        pass

    def compute_m_t(self, current_segment, past_next_changepoint, meta_name_tensor_one_hot):
        """m_t represents the value at the origin(t=0) that we would need to have so that if we use (k_t + k_0) as
        slope, we reach the same value at time = chagepoint_start_of_segment_i as we would reach by following the
        segmented slope (having in each segment the slope trend_deltas(i) + k_0)

        Parameters
        ----------
            current_segment : torch.Tensor, int
                segment corresponding to time t (batch_size, n_forecasts, segments (+ 1))

            past_next_changepoint : torch.Tensor, bool
                whether the a changepoint >= time t (batch_size, n_forecasts, segments (+ 1))

            meta_name_tensor_one_hot : torch.Tensor, float
                Metadata about the all the samples of the model input batch.

                Contains the following:
                    * ``df_name`` (list, str), time series name ID corresponding to each sample of the input batch.
        Returns
        -------
            torch.Tensor
                m_t,  ( batch_size, n_forecasts, quantiles_size)
        """

        pass

    def compute_trend(self, t, k_t, m_t, meta_name_tensor_one_hot=None):
        """This method computes the trend component of the model.


        Parameters
        ----------
            t : torch.Tensor, float
                time

            k_t : torch.Tensor, int
                see compute_k_t

            m_t : torch.Tensor, bool
                see compute_m_t

            meta_name_tensor_one_hot : torch.Tensor, float
                Metadata about the all the samples of the model input batch.

                Contains the following:
                    * ``df_name`` (list, str), time series name ID corresponding to each sample of the input batch.
        Returns
        -------
            torch.Tensor
                trend_component,  ( batch_size, n_forecasts, quantiles_size)
        """
        pass


class GlobalPiecewiseLinearTrend(PiecewiseLinearTrend):
    def __init__(self, config, id_list, quantiles, num_trends_modelled, n_forecasts, device):
        super().__init__(
            config=config,
            n_forecasts=n_forecasts,
            num_trends_modelled=num_trends_modelled,
            quantiles=quantiles,
            id_list=id_list,
            device=device,
        )

    def compute_k_t(self, current_segment, past_next_changepoint, meta_name_tensor_one_hot=None):
        """This method overrides the method from the PiecewiseLinear class."""
        # For segmentwise k_t, is the model parameter representing the trend slope(actually, trend slope-k_0) in the
        # current_segment at time t (for each sample of the batch).
        # k_t = k_t(current_segment).
        # dimensions - batch_size, n_forecasts, quantiles_size
        k_t = torch.sum(
            current_segment.unsqueeze(dim=2) * self.trend_deltas.permute(1, 0, 2).unsqueeze(1),
            dim=-1,
        )

        # For not segmentwise k_t is the model parameter representing the difference between trend slope in the
        # current_segment at time t and the trend slope in the previous segment (for each sample of the batch).
        if not self.segmentwise_trend:
            # k_t = k_t(current_segment, previous_segment)
            # dimensions - batch_size, n_forecasts, quantiles_size
            previous_deltas_t = torch.sum(
                past_next_changepoint.unsqueeze(dim=2) * self.trend_deltas.permute(1, 0, 2)[:, :, :-1].unsqueeze(dim=0),
                dim=-1,
            )
            k_t = k_t + previous_deltas_t

        return k_t

    def compute_m_t(self, current_segment, past_next_changepoint, meta_name_tensor_one_hot=None):
        """This method overrides the method from the PiecewiseLinear class."""
        # Computing m_t.
        # m_t represents the value at the origin(t=0) that we would need to have so that if we use (k_t + k_0) as slope,
        # we reach the same value at time = chagepoint_start_of_segment_i
        # that the segmented slope (having in each segment the slope trend_deltas(i) + k_0)
        if self.config_trend.growth != "discontinuous":
            # Intermediate computation: deltas.
            # `deltas`` is representing the difference between trend slope in the current_segment at time t
            #  and the trend slope in the previous segment.
            if self.segmentwise_trend:
                # dimensions - quantiles, num_trends_modelled, segments
                deltas = self.trend_deltas[:, :, :] - torch.cat((self.trend_k0, self.trend_deltas[:, :, 0:-1]), dim=2)

            else:
                deltas = self.trend_deltas

            # dimensions - quantiles, 1, segments
            gammas = -self.trend_changepoints_t[1:] * deltas[:, :, 1:]
            # dimensions - batch_size, n_forecasts, quantiles
            m_t = torch.sum(past_next_changepoint.unsqueeze(dim=2) * gammas.permute(1, 0, 2).unsqueeze(1), dim=-1)

            if not self.segmentwise_trend:
                m_t = m_t.detach()
        else:
            # For discontinuous, trend_m is a parameter to optimize, as it is not defined just
            # by trend_deltas & trend_k0
            # m_t = m_t(current_segment)
            # dimensions - batch_size, n_forecasts, quantiles
            m_t = torch.sum(current_segment.unsqueeze(dim=2) * self.trend_m.permute(1, 0, 2).unsqueeze(dim=0), dim=-1)

        return m_t

    def compute_trend(self, t, k_t, m_t, meta_name_tensor_one_hot=None):
        """This method overrides the method from the PiecewiseLinear class."""
        # Computing trend value at time(t) for each batch sample.
        # dimensions - batch_size, n_forecasts, quantiles
        return (self.trend_k0.permute(1, 2, 0) + k_t) * torch.unsqueeze(t, dim=2) + m_t


class LocalPiecewiseLinearTrend(PiecewiseLinearTrend):
    def __init__(self, config, id_list, quantiles, num_trends_modelled, n_forecasts, device):
        super().__init__(
            config=config,
            n_forecasts=n_forecasts,
            num_trends_modelled=num_trends_modelled,
            quantiles=quantiles,
            id_list=id_list,
            device=device,
        )

    def compute_k_t(self, current_segment, past_next_changepoint, meta_name_tensor_one_hot):
        """This method overrides the method from the PiecewiseLinear class."""
        # For segmentwise k_t, is the model parameter representing the trend slope(actually, trend slope-k_0) in the
        # current_segment at time t (for each sample of the batch).
        # k_t = k_t(current_segment, sample metadata)
        # dimensions - quantiles, batch_size, segments (+ 1)
        trend_deltas_by_sample = torch.sum(
            meta_name_tensor_one_hot.unsqueeze(dim=0).unsqueeze(dim=-1) * self.trend_deltas.unsqueeze(dim=1), dim=2
        )
        # dimensions - batch_size, n_forecasts, quantiles_size
        k_t = torch.sum(current_segment.unsqueeze(dim=2) * trend_deltas_by_sample.permute(1, 0, 2).unsqueeze(1), dim=-1)

        # For not segmentwise k_t is the model parameter representing the difference between trend slope in the
        # current_segment at time t and the trend slope in the previous segment (for each sample of the batch).
        if not self.segmentwise_trend:
            # k_t = k_t(current_segment, previous_segment, sample metadata)
            previous_deltas_t = torch.sum(
                past_next_changepoint.unsqueeze(dim=2)
                * trend_deltas_by_sample.permute(1, 0, 2)[:, :, :-1].unsqueeze(dim=1),
                dim=-1,
            )
            # dimensions - batch_size, n_forecasts, quantiles_size
            k_t = k_t + previous_deltas_t

        return k_t

    def compute_m_t(self, current_segment, past_next_changepoint, meta_name_tensor_one_hot=None):
        """This method overrides the method from the PiecewiseLinear class."""
        # Computing m_t.
        # m_t represents the value at the origin(t=0) that we would need to have so that if we use (k_t + k_0) as slope,
        # we reach the same value at time = chagepoint_start_of_segment_i
        # that the segmented slope (having in each segment the slope trend_deltas(i) + k_0)
        if self.config_trend.growth != "discontinuous":
            # Intermediate computation: deltas.
            # `deltas`` is representing the difference between trend slope in the current_segment at time t
            #  and the trend slope in the previous segment.
            if self.segmentwise_trend:
                # dimensions - quantiles, num_trends_modelled, segments
                deltas = self.trend_deltas[:, :, :] - torch.cat((self.trend_k0, self.trend_deltas[:, :, 0:-1]), dim=2)

            else:
                deltas = self.trend_deltas

            # We create a dict of gammas based on the df_name
            # m_t = m_t(current_segment, sample metadata)
            # dimensions - quantiles, num_time_series, segments
            gammas_0 = -self.trend_changepoints_t[1:] * deltas[:, :, 1:]
            # dimensions - quantiles, segments, batch_size
            gammas = torch.sum(
                torch.transpose(meta_name_tensor_one_hot, 1, 0).unsqueeze(dim=-2).unsqueeze(dim=0)
                * torch.unsqueeze(gammas_0, dim=-1),
                dim=1,
            )
            # dimensions - batch_size, n_forecasts, quantiles
            m_t = torch.sum(past_next_changepoint.unsqueeze(2) * gammas.permute(2, 0, 1).unsqueeze(1), dim=-1)

            if not self.segmentwise_trend:
                m_t = m_t.detach()
        else:
            # For discontinuous, trend_m is a parameter to optimize, as it is not defined just
            # by trend_deltas & trend_k0
            # m_t = m_t(current_segment, sample metadata)
            # dimensions - quantiles, batch_size, segments
            m_t_0 = torch.sum(
                meta_name_tensor_one_hot.unsqueeze(dim=0).unsqueeze(dim=-1) * self.trend_m.unsqueeze(dim=1), dim=2
            )
            # dimensions - batch_size, n_forecasts, quantiles
            m_t = torch.sum(
                current_segment.unsqueeze(dim=2) * m_t_0.permute(1, 0, 2).unsqueeze(dim=1),
                dim=-1,
            )

        return m_t

    def compute_trend(self, t, k_t, m_t, meta_name_tensor_one_hot=None):
        """This method overrides the method from the PiecewiseLinear class."""
        # Computing trend value at time(t) for each batch sample.
        # trend_k_0 = trend_k_0(current_segment, sample metadata)
        trend_k_0 = torch.sum(
            meta_name_tensor_one_hot.unsqueeze(dim=0).unsqueeze(dim=-1) * self.trend_k0.unsqueeze(dim=1), dim=2
        ).permute(1, 2, 0)
        # dimensions - batch_size, n_forecasts, quantiles
        return (trend_k_0 + k_t) * t.unsqueeze(dim=2) + m_t



import os

import numpy as np
import pandas as pd


def read_json(path, metrics_path, branch):
    try:
        df = pd.read_json(os.path.join(metrics_path, path), orient="index")
        df = df.reset_index()
        df.rename(columns={0: branch, "index": "Metric"}, inplace=True)
        return df
    except:  # noqa: E722
        return None


# Load all metrics paths
metrics_path = os.path.join(os.getcwd(), "tests", "metrics")
metrics_path_main = os.path.join(os.getcwd(), "tests", "metrics-main")
metrics_files = [f for f in os.listdir(metrics_path) if ".json" in f]
all_metrics = pd.DataFrame()

# Loop through all metrics files
for f in metrics_files:
    current = read_json(f, metrics_path, "current")
    main = read_json(f, metrics_path_main, "main")

    if main is not None:
        df = pd.merge(current, main, on=["Metric"], how="left")

        # Adjust the runtime metrics
        time = df.loc[df["Metric"] == "time"]
        performance = df.loc[df["Metric"] == "system_performance"]
        adjusted_performance = time["main"].values[0] / performance["main"].values[0] * performance["current"].values[0]
        df.loc[df["Metric"] == "time", "main"] = adjusted_performance

        df["diff"] = ((df["main"] - df["current"]) / df["main"]) * 100 * -1

        df = df.round(5)
        df[" "] = np.select(
            condlist=[(df["diff"] >= 7.0), (df["diff"] >= 3.0), (df["diff"] <= -5.0)],
            choicelist=[":x:", ":warning:", ":tada:"],
            default=":white_check_mark:",
        )
        df["diff"] = df["diff"].fillna(0)
        df["diff"] = df["diff"] + 0.0
        df["diff"] = df["diff"].round(2).astype(str) + "%"
    else:
        df = current.copy()
        df["main"] = "-"
        df["diff"] = "-"
        df[" "] = ""
        df = df.round(4)

    # Remove unused metrics
    df = df[~df["Metric"].isin(["epoch", "RegLoss", "RegLoss_val", "system_performance", "system_std"])]
    df["Benchmark"] = f.split(".")[0]
    df = df[["Benchmark", "Metric", "main", "current", "diff", " "]]
    all_metrics = pd.concat([all_metrics, df])

print(str(all_metrics.to_markdown(tablefmt="github", index=False)))



#!/usr/bin/env python3

import holidays
import pytest

from neuralprophet import hdays_utils


def test_get_country_holidays():
    assert issubclass(hdays_utils.get_country_holidays("TU").__class__, holidays.countries.turkey.TR) is True

    for country in ("UnitedStates", "US", "USA"):
        us_holidays = hdays_utils.get_country_holidays(country, years=2019)
        assert issubclass(us_holidays.__class__, holidays.countries.united_states.UnitedStates) is True
        assert len(us_holidays) == 10

    with pytest.raises(AttributeError):
        hdays_utils.get_country_holidays("NotSupportedCountry")



from neuralprophet.components.trend import Trend


class StaticTrend(Trend):
    def __init__(self, config, id_list, quantiles, num_trends_modelled, n_forecasts, device):
        super().__init__(
            config=config,
            n_forecasts=n_forecasts,
            num_trends_modelled=num_trends_modelled,
            quantiles=quantiles,
            id_list=id_list,
            device=device,
        )

    def forward(self, t, meta):
        """
        Computes trend based on model configuration.

        Parameters
        ----------
            t : torch.Tensor float
                normalized time, dim: (batch, n_forecasts)
            meta: dict
                Metadata about the all the samples of the model input batch. Contains the following:
                    * ``df_name`` (list, str), time series ID corresponding to each sample of the input batch.
        Returns
        -------
            torch.Tensor
                Trend component, same dimensions as input t
        """
        return self.bias.unsqueeze(dim=0).repeat(t.shape[0], t.shape[1], 1)

    @property
    def get_trend_deltas(self):
        pass

    def add_regularization(self):
        pass
